from __future__ import annotations

import curses
from dataclasses import dataclass, replace
from typing import Generic, Self, override

from tjex.config import config
from tjex.curses_helper import DummyRegion, OffsetRegion, Region, SubRegion
from tjex.panel import Event, KeyBindings, KeyPress, Panel
from tjex.point import Point
from tjex.table import T_Cell, T_Key, Table
from tjex.utils import TjexError


class TableEmptyError(TjexError):
    def __init__(self):
        super().__init__("Table is empty")


@dataclass
class TableState:
    cursor: Point
    content_base: Point

    @property
    def row_only(self):
        return TableState(replace(self.cursor, x=0), replace(self.content_base, x=0))

    @property
    def col_only(self):
        return TableState(replace(self.cursor, y=0), replace(self.content_base, y=0))


class TablePanel(Panel, Generic[T_Key, T_Cell]):
    bindings: KeyBindings[Self, Event | None] = KeyBindings()

    def __init__(self, region: Region):
        self._max_cell_width: int | None = None
        self.full_cell_width: bool = False
        self.table: Table[T_Key, T_Cell] = Table({}, [], [], [], [], [])
        self.offsets: list[int] = []
        self.content_offset: Point = Point(1, 0)
        self.cursor: Point = Point(0, 0)
        self.region: Region = region
        self.content_region: OffsetRegion = OffsetRegion(DummyRegion(), Point.ZERO)
        self.row_header_region: OffsetRegion = OffsetRegion(DummyRegion(), Point.ZERO)
        self.col_header_region: OffsetRegion = OffsetRegion(DummyRegion(), Point.ZERO)

    @override
    def resize(self, region: Region):
        self.region = region
        self.row_header_region = OffsetRegion(
            SubRegion(
                region,
                replace(self.content_offset, x=0),
                region.size - replace(self.content_offset, x=0),
            ),
            self.row_header_region.offset,
        )
        self.col_header_region = OffsetRegion(
            SubRegion(
                region,
                replace(self.content_offset, y=0),
                region.size - replace(self.content_offset, y=0),
            ),
            self.col_header_region.offset,
        )
        self.content_region = OffsetRegion(
            SubRegion(region, self.content_offset, region.size - self.content_offset),
            self.content_region.offset,
        )

    @property
    def max_cell_width(self):
        if self.full_cell_width:
            return None
        return self._max_cell_width or config.max_cell_width

    def update(self, table: Table[T_Key, T_Cell], state: TableState | None):
        self.table = table

        self.offsets = [0]
        for formatter in table.col_formatters[1:]:
            self.offsets.append(
                self.offsets[-1] + formatter.final_width(self.max_cell_width) + 1
            )

        if not table.col_formatters:
            return

        self.content_offset = Point(
            1,
            table.col_formatters[0].final_width(self.max_cell_width) + 1,
        )
        self.resize(self.region)
        if state is not None:
            self.cursor = state.cursor
            self.content_region.offset = state.content_base
            self.clamp_cursor()
        else:
            self.cursor = Point(0, 0)
            self.content_region.offset = Point(0, 0)
        self.propagate_content_base()

    @property
    def state(self):
        return TableState(self.cursor, self.content_region.offset)

    @override
    def draw(self):
        table = self.table

        col_range = range(
            max(
                0,
                next(
                    (
                        i
                        for i, offset in enumerate(self.offsets)
                        if offset > self.content_region.offset.x
                    ),
                    len(self.offsets),
                )
                - 1,
            ),
            next(
                (
                    i
                    for i, offset in enumerate(self.offsets)
                    if offset
                    >= self.content_region.offset.x + self.content_region.width
                ),
                len(self.offsets) - 1,
            ),
        )
        row_range = range(
            self.content_region.offset.y,
            min(
                self.content_region.offset.y + self.content_region.height,
                len(table.row_keys),
            ),
        )

        for i in col_range:
            table.col_formatters[i + 1].draw(
                table.col_headers[i],
                self.col_header_region,
                Point(0, self.offsets[i]),
                self.max_cell_width,
                curses.A_BOLD,
                True,
            )

        for i in row_range:
            table.col_formatters[0].draw(
                table.row_headers[i],
                self.row_header_region,
                Point(i, 0),
                self.max_cell_width,
                curses.A_BOLD,
                False,
            )

        for i in row_range:
            for j in col_range:
                cell = table.content[table.row_keys[i]].get(table.col_keys[j], None)
                if cell is not None:
                    table.col_formatters[j + 1].draw(
                        cell,
                        self.content_region,
                        Point(i, self.offsets[j]),
                        self.max_cell_width,
                        0,
                        False,
                    )

        if self.active:
            self.chgat_cursor(curses.A_REVERSE)

    def chgat_cursor(self, a: int):
        if self.table.col_keys:
            self.content_region.chgat(
                Point(self.cursor.y, self.offsets[self.cursor.x]),
                self.table.col_formatters[self.cursor.x + 1].final_width(
                    self.max_cell_width
                ),
                a,
            )

    def clamp_cursor(self):
        table = self.table
        self.cursor = Point(
            max(0, min(len(table.row_keys) - 1, self.cursor.y)),
            max(0, min(len(table.col_keys) - 1, self.cursor.x)),
        )
        if not table.col_keys:
            return

        if (
            self.offsets[self.cursor.x + 1]
            >= self.content_region.width + self.content_region.offset.x
        ):
            self.content_region.offset = replace(
                self.content_region.offset,
                x=self.offsets[self.cursor.x + 1] - 1 - self.content_region.width,
            )
        if self.offsets[self.cursor.x] < self.content_region.offset.x:
            self.content_region.offset = replace(
                self.content_region.offset, x=self.offsets[self.cursor.x]
            )

        if (
            self.cursor.y + 1
            > self.content_region.height + self.content_region.offset.y
        ):
            self.content_region.offset = replace(
                self.content_region.offset,
                y=self.cursor.y + 1 - self.content_region.height,
            )
        if self.cursor.y < self.content_region.offset.y:
            self.content_region.offset = replace(
                self.content_region.offset, y=self.cursor.y
            )

        self.propagate_content_base()

    def propagate_content_base(self):
        self.row_header_region.offset = replace(self.content_region.offset, x=0)
        self.col_header_region.offset = replace(self.content_region.offset, y=0)

    @bindings.add("KEY_UP", "p", "C-p")
    def up(self):
        self.cursor += Point(-1, 0)

    @bindings.add("KEY_DOWN", "n", "C-n")
    def down(self):
        self.cursor += Point(1, 0)

    @bindings.add("KEY_LEFT", "b", "C-b")
    def left(self):
        self.cursor += Point(0, -1)

    @bindings.add("KEY_RIGHT", "f", "C-f")
    def right(self):
        self.cursor += Point(0, 1)

    @property
    def row_key(self):
        if not self.table.row_keys:
            raise TableEmptyError()
        return self.table.row_keys[self.cursor.y]

    @property
    def col_key(self):
        if not self.table.col_keys:
            raise TableEmptyError()
        return self.table.col_keys[self.cursor.x]

    @property
    def cell_keys(self):
        return (self.row_key, self.col_key)

    @bindings.add("M-<")
    def first_row(self):
        """Jump to first row"""
        self.cursor = replace(self.cursor, y=0)

    @bindings.add("M->")
    def last_row(self):
        """Jump to last row"""
        self.cursor = replace(self.cursor, y=len(self.table.row_keys) - 1)

    @bindings.add("KEY_NPAGE")
    def next_page(self):
        self.cursor += Point(self.content_region.height, 0)

    @bindings.add("KEY_PPAGE")
    def prev_page(self):
        self.cursor -= Point(self.content_region.height, 0)

    @bindings.add("KEY_END", "C-e")
    def last_col(self):
        """Jump to last column"""
        self.cursor = replace(self.cursor, x=len(self.table.col_keys) - 1)

    @bindings.add("KEY_HOME", "C-a")
    def first_col(self):
        """Jump to first column"""
        self.cursor = replace(self.cursor, x=0)

    @bindings.add("l")
    def full_width(self):
        """Toggle: rendering all cells with their full width vs. max_cell_width"""
        self.full_cell_width = not self.full_cell_width
        self.update(self.table, self.state)

    @bindings.add("+")
    def inc_width(self):
        """Increase max_cell_width by one"""
        self._max_cell_width = (self._max_cell_width or config.max_cell_width) + 1
        self.update(self.table, self.state)

    @bindings.add("-")
    def dec_width(self):
        """Decrease max_cell_width by one"""
        self._max_cell_width = max(
            1, (self._max_cell_width or config.max_cell_width) - 1
        )
        self.update(self.table, self.state)

    @override
    def handle_key(self, key: KeyPress) -> list[Event]:
        res = self.bindings.handle_key(key, self)
        if isinstance(res, str):
            return [res]
        self.clamp_cursor()
        if res is None:
            return []
        return [res]



import json
import subprocess as sp
import tomllib
from base64 import b64encode
from dataclasses import asdict, dataclass, field
from pathlib import Path
from textwrap import dedent
from typing import Any, TypeVar

from tjex.curses_helper import KEY_ALIASES
from tjex.panel import KeyBindings
from tjex.utils import TjexError


class ConfigError(TjexError):
    pass


T = TypeVar("T")


def config_field(default: T, doc: str) -> T:
    return field(default=default, metadata={"doc": dedent(doc).strip()})


@dataclass
class Config:
    bindings: dict[str, dict[str, str]] = field(default_factory=dict)
    float_precision: int = config_field(
        8, "Number of significand digits for floating point numbers"
    )
    max_cell_width: int = config_field(50, "Default maximum cell width")
    copy_command: str | None = config_field(
        None,
        """
        Bash command to copy a string from stdin to the clipboard
        e.g.
        * "osc" (https://github.com/theimpostor/osc)
        * "wl-copy 2> /dev/null"
          (It's important to pipe stderr to null, because `wl-copy` will stick around and make tjex hang otherwise)
        By default, tjex just emits the plain OCS52 escape sequence.
        For many setups, this wont work though.
        """,
    )
    append_history_command: str = config_field(
        "atuin history end --exit 0 -- $(atuin history start -- {})",
        "Format string that generates a bash command to append the given string to a shell history",
    )
    start_at_prompt: bool = config_field(
        False, "On startup, focus the jq prompt instead of the table panel"
    )
    jq_command: str = config_field("jq", "Name of or full path to the jq executable")

    def do_copy(self, s: str):
        if self.copy_command is None:
            print("\033]52;c;{}\a".format(b64encode(s.encode()).decode()))
        else:
            result = sp.run(
                ["bash", "-c", self.copy_command],
                input=s + "\n",
                capture_output=True,
                text=True,
            )
            if result.returncode != 0:
                raise ConfigError(result.stderr)


config: Config = Config()


def comment_out(s: str):
    return "".join(f"# {l}\n" if l else "\n" for l in s.splitlines())


def make_example_config(bindings: dict[str, KeyBindings[Any, Any]]):
    res = "# Example configuration for tjex"

    for k, v in asdict(config).items():
        if k != "bindings":
            res += "\n\n"
            if doc := config.__dataclass_fields__[k].metadata.get("doc"):
                res += comment_out(doc)
            if v is None:
                res += comment_out(f'{json.dumps(k)} = ""')
            else:
                res += f"{json.dumps(k)} = {json.dumps(v)}"

    reverse_aliases = {v: k for k, v in KEY_ALIASES.items()}
    for panel, b in bindings.items():
        res += f"\n\n"
        res += f"[bindings.{panel}]"
        for f in b.functions:
            res += f"\n"
            res += comment_out(f.name)
            if f.description is not None:
                res += comment_out(f.description)
            for k, v in b.bindings.items():
                if v == f:
                    res += f"{json.dumps(reverse_aliases.get(k, k))} = {json.dumps(f.name)}\n"

    return comment_out(res)


def apply_bindings(bindings: dict[str, KeyBindings[Any, Any]]):
    for panel, b in config.bindings.items():
        for k, f in b.items():
            bindings[panel].bindings[KEY_ALIASES.get(k, k)] = next(
                _f for _f in bindings[panel].functions if _f.name == f
            )


def load(
    config_file: Path,
    bindings: dict[str, KeyBindings[Any, Any]],
) -> None:
    if config_file.exists():
        with open(config_file, "rb") as f:
            for k, v in tomllib.load(f).items():
                setattr(config, k, v)
        apply_bindings(bindings)
    else:
        config_file.parent.mkdir(parents=True, exist_ok=True)
        with open(config_file, "w") as f:
            _ = f.write(make_example_config(bindings))



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterable
from dataclasses import dataclass
from typing import Callable, Generic, TypeVar

from tjex.curses_helper import KEY_ALIASES, Region


class Event(ABC):
    pass


@dataclass
class StatusUpdate(Event):
    msg: str


T = TypeVar("T")
S = TypeVar("S")
R = TypeVar("R")


@dataclass(frozen=True)
class KeyPress(Event):
    key: str


@dataclass
class Function(Generic[T, S]):
    f: Callable[[T], S]
    name: str
    description: str | None


class KeyBindings(Generic[T, S]):
    def __init__(self):
        self.functions: list[Function[T, S]] = []
        self.bindings: dict[str, Function[T, S]] = {}

    def add(self, *key: str, name: str | None = None):
        def wrap(f: Callable[[T], S]) -> Callable[[T], S]:
            func = Function(f, f.__name__ if name is None else name, f.__doc__)
            self.functions.append(func)
            for k in key:
                self.bindings[KEY_ALIASES.get(k, k)] = func
            return f

        return wrap

    def handle_key(self, key: KeyPress | R, p: T) -> KeyPress | R | S:
        if isinstance(key, KeyPress) and key.key in self.bindings:
            return self.bindings[key.key].f(p)
        return key


class Panel(ABC):
    active: bool = False

    @abstractmethod
    def resize(self, region: Region):
        pass

    @abstractmethod
    def handle_key(self, key: KeyPress) -> Iterable[Event]:
        pass

    @abstractmethod
    def draw(self):
        pass

    def set_active(self, active: bool):
        self.active = active



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, Generic, TypeVar

from tjex.point import Point

if TYPE_CHECKING:
    from tjex.curses_helper import Region

T_Cell = TypeVar("T_Cell")
T_Key = TypeVar("T_Key")


class CellFormatter(ABC, Generic[T_Cell]):
    min_width: int
    width: int

    @abstractmethod
    def draw(
        self,
        cell: T_Cell,
        region: Region,
        pos: Point,
        max_width: int | None,
        attr: int,
        force_left: bool,
    ) -> None: ...

    def final_width(self, max_width: int | None):
        width = self.width
        if max_width is not None:
            width = min(width, max_width)
        return max(width, self.min_width)


@dataclass
class Table(Generic[T_Key, T_Cell]):
    content: dict[T_Key, dict[T_Key, T_Cell]]
    row_keys: list[T_Key]
    row_headers: list[T_Cell]
    col_keys: list[T_Key]
    col_headers: list[T_Cell]
    col_formatters: list[CellFormatter[T_Cell]]



from __future__ import annotations

from dataclasses import dataclass
from typing import ClassVar


@dataclass(frozen=True)
class Point:
    y: int
    x: int

    def __add__(self, that: Point):
        return Point(self.y + that.y, self.x + that.x)

    def __sub__(self, that: Point):
        return Point(self.y - that.y, self.x - that.x)

    ZERO: ClassVar[Point]


Point.ZERO = Point(0, 0)



from __future__ import annotations

import curses
from abc import ABC, abstractmethod
from dataclasses import dataclass, replace
from typing import override

from tjex.logging import logger
from tjex.point import Point


def setup_plain_colors():
    curses.start_color()
    for i in range(1, 8):
        curses.init_pair(i, i, curses.COLOR_BLACK)


KEY_ALIASES = {
    "C-g": "\x07",
    "C-d": "\x04",
    "C-o": "\x0f",
    "C-_": "\x1f",
    "C-e": "\x05",
    "C-a": "\x01",
    "C-k": "\x0b",
    "C-<delete>": "kDC5",
    "C-<backspace>": "\x08",
    "C-f": "\x06",
    "C-<right>": "kRIT5",
    "M-<right>": "kRIT3",
    "C-b": "\x02",
    "C-<left>": "kLFT5",
    "M-<left>": "kLFT3",
    "C-p": "\x10",
    "C-n": "\x0e",
    "C-y": "\x19",
}


class KeyReader:
    def __init__(self, window: curses.window):
        curses.set_escdelay(10)
        window.keypad(True)
        window.nodelay(True)
        window.notimeout(False)
        self.window: curses.window = window

    def get(self) -> str | None:
        try:
            prefix = ""
            key = self.window.get_wch()
            if key == "\x1b":
                try:
                    prefix = "M-"
                    key = self.window.get_wch()
                except curses.error:
                    return "ESC"
            if isinstance(key, int):
                key = curses.keyname(key).decode("utf-8")
            key = prefix + key
            logger.debug(f"{key=}")
            return key
        except curses.error:
            return None


class Region(ABC):
    size: Point = Point.ZERO

    @property
    def height(self):
        return self.size.y

    @property
    def width(self):
        return self.size.x

    @abstractmethod
    def insstr(self, pos: Point, s: str, attr: int = 0) -> None:
        pass

    @abstractmethod
    def chgat(self, pos: Point, width: int, attr: int) -> None:
        pass

    def resize(self):
        """If this region's size depends on external resources, update it.
        (Currently only used for WindowRegion)
        """


class DummyRegion(Region):
    def __init__(self, size: Point = Point.ZERO):
        self.size: Point = size
        self.content: list[str] = self.height * [self.width * " "]

    def clear(self):
        self.content = self.height * [self.width * " "]

    @override
    def insstr(self, pos: Point, s: str, attr: int = 0) -> None:
        self.content[pos.y] = (
            self.content[pos.y][: pos.x] + s + self.content[pos.y][pos.x :]
        )[: self.width]

    @override
    def chgat(self, pos: Point, width: int, attr: int) -> None:
        pass


class WindowRegion(Region):
    def __init__(self, window: curses.window):
        self.window: curses.window = window
        self.resize()

    @override
    def insstr(self, pos: Point, s: str, attr: int = 0):
        if self.height > pos.y >= 0 and self.width > pos.x > -len(s):
            self.window.insstr(
                pos.y, max(0, pos.x), s[max(0, -pos.x) : self.width - pos.x], attr
            )

    @override
    def chgat(self, pos: Point, width: int, attr: int):
        if self.height > pos.y >= 0 and self.width > pos.x > -width:
            self.window.chgat(
                pos.y, max(0, pos.x), min(self.width, width - max(0, -pos.x)), attr
            )

    @override
    def resize(self):
        self.size: Point = Point(*self.window.getmaxyx())


@dataclass
class SubRegion(Region):
    parent: Region
    pos: Point = Point.ZERO
    size: Point = Point.ZERO

    @property
    def y(self):
        return self.pos.y

    @property
    def x(self):
        return self.pos.x

    @override
    def insstr(self, pos: Point, s: str, attr: int = 0):
        if self.height > pos.y >= 0 and self.width > pos.x > -len(s):
            self.parent.insstr(
                self.pos + replace(pos, x=max(0, pos.x)),
                s[max(0, -pos.x) : self.width - pos.x],
                attr,
            )

    @override
    def chgat(self, pos: Point, width: int, attr: int):
        if self.height > pos.y >= 0 and self.width > pos.x > -width:
            self.parent.chgat(
                self.pos + replace(pos, x=max(0, pos.x)),
                min(self.width, width - max(0, -pos.x)),
                attr,
            )


class OffsetRegion(Region):
    def __init__(self, parent: Region, offset: Point):
        self.parent: Region = parent
        self.size: Point = parent.size
        self.offset: Point = offset

    @override
    def insstr(self, pos: Point, s: str, attr: int = 0):
        self.parent.insstr(pos - self.offset, s, attr)

    @override
    def chgat(self, pos: Point, width: int, attr: int):
        self.parent.chgat(pos - self.offset, width, attr)



from __future__ import annotations

import logging
from pathlib import Path

logger = logging.getLogger("tjex")


def setup(logfile: Path | None):
    if logfile is not None:
        logging.basicConfig(
            filename=logfile.as_posix(),
            level=logging.DEBUG,
        )



from __future__ import annotations

from typing import Generic, TypeVar

T = TypeVar("T")


class History(Generic[T]):
    history: list[T]
    position: int

    @property
    def current(self):
        return self.history[self.position]

    def __init__(self, default: T):
        self.history = [default]
        self.position = 0

    def push(self, v: T):
        if self.current != v:
            self.history = [*self.history[: self.position + 1], v]
            self.position += 1

    def pop(self, current: T):
        if current != self.current:
            self.push(current)
        self.position = max(0, self.position - 1)
        return self.current

    def redo(self):
        self.position = min(self.position + 1, len(self.history) - 1)
        return self.current
